"""Core functionality for `onprem`"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_core.ipynb.

# %% auto 0
__all__ = ['MIN_MODEL_SIZE', 'MISTRAL_MODEL_URL', 'MISTRAL_MODEL_ID', 'MISTRAL_PROMPT_TEMPLATE', 'ZEPHYR_MODEL_URL',
           'ZEPHYR_MODEL_ID', 'ZEPHYR_PROMPT_TEMPLATE', 'LLAMA_MODEL_URL', 'LLAMA_MODEL_ID', 'LLAMA_PROMPT_TEMPLATE',
           'MODEL_URL_DICT', 'MODEL_ID_DICT', 'LLAMA_CPP', 'TRANSFORMERS', 'ENGINE_DICT', 'PROMPT_DICT',
           'DEFAULT_MODEL', 'DEFAULT_ENGINE', 'DEFAULT_EMBEDDING_MODEL', 'DEFAULT_QA_PROMPT',
           'AnswerConversationBufferMemory', 'LLM']

# %% ../nbs/00_core.ipynb 3
from . import utils as U
from langchain.chains import RetrievalQA, ConversationalRetrievalChain
from langchain.memory import ConversationBufferMemory
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.prompts import PromptTemplate
from langchain_community.llms import LlamaCpp
from langchain_openai import ChatOpenAI, AzureChatOpenAI
import os
import warnings
from typing import Any, Dict, Optional, Callable, Union, List

# %% ../nbs/00_core.ipynb 4
# reference: https://github.com/langchain-ai/langchain/issues/5630#issuecomment-1574222564
class AnswerConversationBufferMemory(ConversationBufferMemory):
    def save_context(self, inputs: Dict[str, Any], outputs: Dict[str, str]) -> None:
        return super(AnswerConversationBufferMemory, self).save_context(
            inputs, {"response": outputs["answer"]}
        )

MIN_MODEL_SIZE = 250000000
MISTRAL_MODEL_URL = "https://huggingface.co/TheBloke/Mistral-7B-Instruct-v0.2-GGUF/resolve/main/mistral-7b-instruct-v0.2.Q4_K_M.gguf"
MISTRAL_MODEL_ID = "TheBloke/Mistral-7B-Instruct-v0.2-AWQ"
MISTRAL_PROMPT_TEMPLATE = "[INST] {prompt} [/INST]"
ZEPHYR_MODEL_URL = "https://huggingface.co/TheBloke/zephyr-7B-beta-GGUF/resolve/main/zephyr-7b-beta.Q4_K_M.gguf"
ZEPHYR_MODEL_ID = "TheBloke/zephyr-7B-beta-AWQ"
ZEPHYR_PROMPT_TEMPLATE = "<|system|>\n</s>\n<|user|>\n{prompt}</s>\n<|assistant|>"
LLAMA_MODEL_URL = "https://huggingface.co/lmstudio-community/Meta-Llama-3.1-8B-Instruct-GGUF/resolve/main/Meta-Llama-3.1-8B-Instruct-Q4_K_M.gguf"
LLAMA_MODEL_ID = "hugging-quants/Meta-Llama-3.1-8B-Instruct-AWQ-INT4"
LLAMA_PROMPT_TEMPLATE = """<|start_header_id|>system<|end_header_id|>

You are a super-intelligent helpful assistant that executes instructions.<|eot_id|><|start_header_id|>user<|end_header_id|>

{prompt}<|eot_id|><|start_header_id|>assistant<|end_header_id|>
"""
MODEL_URL_DICT = {'mistral' : MISTRAL_MODEL_URL, 'zephyr':ZEPHYR_MODEL_URL, 'llama' : LLAMA_MODEL_URL}
MODEL_ID_DICT = {'mistral': MISTRAL_MODEL_ID, 'zephyr' : ZEPHYR_MODEL_ID, 'llama': LLAMA_MODEL_ID}
LLAMA_CPP = 'llama.cpp'
TRANSFORMERS = 'transformers'
ENGINE_DICT = {LLAMA_CPP: MODEL_URL_DICT, TRANSFORMERS : MODEL_ID_DICT}
PROMPT_DICT = {'mistral': MISTRAL_PROMPT_TEMPLATE, 'zephyr' : ZEPHYR_PROMPT_TEMPLATE, 'llama' : LLAMA_PROMPT_TEMPLATE} 
DEFAULT_MODEL = 'mistral'
DEFAULT_ENGINE = LLAMA_CPP
DEFAULT_EMBEDDING_MODEL = "sentence-transformers/all-MiniLM-L6-v2"
DEFAULT_QA_PROMPT = """"Use the following pieces of context delimited by three backticks to answer the question at the end. If you don't know the answer, just say that you don't know, don't try to make up an answer.

```{context}```

Question: {question}
Helpful Answer:"""


class LLM:
    def __init__(
        self,
        model_url:Optional[str] = None,
        model_id:Optional[str] = None,
        default_model:str  = DEFAULT_MODEL,
        default_engine:str = DEFAULT_ENGINE,
        n_gpu_layers: Optional[int] = None,
        prompt_template: Optional[str] = None,
        model_download_path: Optional[str] = None,
        vectordb_path: Optional[str] = None,
        max_tokens: int = 512,
        n_ctx: int = 3900,
        n_batch: int = 1024,
        stop:list=[],
        mute_stream: bool = False,
        callbacks=[],
        embedding_model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
        embedding_model_kwargs: dict = {"device": "cpu"},
        embedding_encode_kwargs: dict = {"normalize_embeddings": False},
        rag_num_source_docs: int = 4,
        rag_score_threshold: float = 0.0,
        check_model_download:bool=True,
        confirm: bool = True,
        verbose: bool = True,
        **kwargs,
    ):
        """
        LLM Constructor.  Extra `kwargs` (e.g., temperature) are fed directly
        to `langchain.llms.LlamaCpp`  (if `model_url` is supplied) or
        `transformers.pipeline` (if `model_id` is supplied).

        **Args:**

        - *model_url*: URL to `.GGUF` model (or the filename if already been downloaded to `model_download_path`).
                       To use an OpenAI-compatible REST API (e.g., vLLM, OpenLLM, Ollama), supply the URL (e.g., `http://localhost:8080/v1`).
                       To use a cloud-based OpenAI model, replace URL with: `openai://<name_of_model>` (e.g., `openai://gpt-3.5-turbo`).
                       To use Azure OpenAI, replace URL with: with: `azure://<deployment_name>`.
                       If None, use the model indicated by `default_model`.
        - *model_id*: Name of or path to Hugging Face model (e.g., in SafeTensor format). Hugging Face Transformers is used for LLM generation instead of **llama-cpp-python**. Mutually-exclusive with `model_url` and `default_model`. The `n_gpu_layers` and `model_download_path` parameters are ignored if `model_id` is supplied.
        - *default_model*: One of {'mistral', 'zephyr', 'llama'}, where mistral is Mistral-Instruct-7B-v0.2, zephyr is Zephyr-7B-beta, and llama is Llama-3.1-8B.
        - *default_engine*: The engine used to run the `default_model`. One of {'llama.cpp', 'transformers'}.
        - *n_gpu_layers*: Number of layers to be loaded into gpu memory. Default is `None`.
        - *prompt_template*: Optional prompt template (must have a variable named "prompt"). Prompt templates are not typically needed when using the `model_id` parameter, as transformers sets it automatically.
        - *model_download_path*: Path to download model. Default is `onprem_data` in user's home directory.
        - *vectordb_path*: Path to vector database (created if it doesn't exist).
                           Default is `onprem_data/vectordb` in user's home directory.
        - *max_tokens*: The maximum number of tokens to generate.
        - *n_ctx*: Token context window. (Llama2 models have max of 4096.)
        - *n_batch*: Number of tokens to process in parallel.
        - *stop*: a list of strings to stop generation when encountered (applied to all calls to `LLM.prompt`)
        - *mute_stream*: Mute ChatGPT-like token stream output during generation
        - *callbacks*: Callbacks to supply model
        - *embedding_model_name*: name of sentence-transformers model. Used for `LLM.ingest` and `LLM.ask`.
        - *embedding_model_kwargs*: arguments to embedding model (e.g., `{device':'cpu'}`).
        - *embedding_encode_kwargs*: arguments to encode method of
                                     embedding model (e.g., `{'normalize_embeddings': False}`).
        - *rag_num_source_docs*: The maximum number of documents retrieved and fed to `LLM.ask` and `LLM.chat` to generate answers
        - *rag_score_threshold*: Minimum similarity score for source to be considered by `LLM.ask` and `LLM.chat`
        - *confirm*: whether or not to confirm with user before downloading a model
        - *verbose*: Verbosity
        """
        self.model_id = None
        self.model_url = None
        if model_url and model_id:
            raise ValueError('The parameters model_url and model_id are mutually-exclusive.')
        elif model_id:
            self.model_id = model_id
            self.model_name = os.path.basename(model_id)
        elif model_url:
            self.model_url = model_url.split("?")[0]
            self.model_name = os.path.basename(model_url)
            self.model_url = MODEL_DICT[default_model] if not model_url else model_url
        else: # neither supplied so use defaults
            url_or_id = ENGINE_DICT[default_engine][default_model]
            self.model_name = os.path.basename(url_or_id)
            if default_engine == LLAMA_CPP:
                self.model_url = url_or_id
                self.model_id = None
                prompt_template = PROMPT_DICT[default_model] if not prompt_template else prompt_template
            else:
                self.model_url = None
                self.model_id = url_or_id
        self.model_download_path = model_download_path or U.get_datadir()

        if self.is_llamacpp():
            try:
                from llama_cpp import Llama
            except ImportError:
                raise ValueError('To run local LLMs, the llama-cpp-python package is required. ' +\
                                 'You can visit https://python.langchain.com/docs/integrations/llms/llamacpp ' +\
                                 'and follow the instructions for your operating system.')
        if self.is_llamacpp() and not os.path.isfile(os.path.join(self.model_download_path, self.model_name)):
            self.download_model(
                self.model_url,
                model_download_path=self.model_download_path,
                confirm=confirm,
            )
        self.prompt_template = prompt_template
        self.vectordb_path = vectordb_path
        self.llm = None
        self.ingester = None
        self.qa = None
        self.chatqa = None
        self.n_gpu_layers = n_gpu_layers
        self.max_tokens = max_tokens
        self.n_ctx = n_ctx
        self.n_batch = n_batch
        self.stop = stop
        self.mute_stream = mute_stream
        self.callbacks = [] if mute_stream else [StreamingStdOutCallbackHandler()]
        if callbacks:
            self.callbacks.extend(callbacks)
        self.embedding_model_name = embedding_model_name
        self.embedding_model_kwargs = embedding_model_kwargs
        self.embedding_encode_kwargs = embedding_encode_kwargs
        self.rag_num_source_docs = rag_num_source_docs
        self.rag_score_threshold = rag_score_threshold
        self.check_model_download = check_model_download
        self.verbose = verbose
        self.extra_kwargs = kwargs


        # explicitly set offload_kqv
        # reference: https://github.com/abetlen/llama-cpp-python/issues/999#issuecomment-1858041458
        # commented out now: no longer needed as of llama-cpp-python==0.2.75
        #self.offload_kqv = True if n_gpu_layers is not None and n_gpu_layers > 0 else False

        # load LLM
        self.load_llm()

        # issue warning
        if self.is_openai_model():
            warnings.warn(f'The model you supplied is {self.model_name}, an external service (i.e., not on-premises). '+\
                          'Use with caution, as your data and prompts will be sent externally.')

    def is_openai_model(self):
        return self.model_url and self.model_url.lower().startswith('openai')

    def is_azure(self):
        return self.model_url and self.model_url.lower().startswith('azure')


    def is_local_api(self):
        basename = os.path.basename(self.model_url) if self.model_url else None
        return self.model_url and self.model_url.lower().startswith('http') and not basename.lower().endswith('.gguf') and not basename.lower().endswith('.bin')

    def is_local(self):
        return not self.is_openai_model() and not self.is_local_api()

    def is_llamacpp(self):
        return self.is_local() and not self.is_hf()

    def is_hf(self):
        return self.model_id is not None

    def update_max_tokens(self, value:int=512):
        """
        Update `max_tokens` (maximum length of generation).
        """
        llm = self.load_llm()
        llm.max_tokens = value


    def update_stop(self, value:list=[]):
        """
        Update `max_tokens` (maximum length of generation).
        """
        llm = self.load_llm()
        llm.stop = value


    @classmethod
    def download_model(
        cls,
        model_url:Optional[str] = None,
        default_model:str=DEFAULT_MODEL,
        model_download_path: Optional[str] = None,
        confirm: bool = True,
        ssl_verify: bool = True,
    ):
        """
        Download an LLM in GGML format supported by [lLama.cpp](https://github.com/ggerganov/llama.cpp).

        **Args:**

        - *model_url*: URL of model. If None, then use default_model.
        - *default_model*: One of {'mistral', 'zephyr', 'llama'}, where mistral is Mistral-Instruct-7B-v0.2, zephyr is Zephyr-7B-beta, and llama is Llama-3.1-8B.
        - *model_download_path*: Path to download model. Default is `onprem_data` in user's home directory.
        - *confirm*: whether or not to confirm with user before downloading
        - *ssl_verify*: If True, SSL certificates are verified.
                        You can set to False if corporate firewall gives you problems.
        """
        model_url = MODEL_URL_DICT[default_model] if not model_url else model_url
        if 'https://huggingface.co' in model_url and 'resolve' not in model_url:
            warnings.warn('\n\nThe supplied URL may not be pointing to the actual GGUF model file.  Please check it.\n\n')
        datadir = model_download_path or U.get_datadir()
        model_name = os.path.basename(model_url)
        filename = os.path.join(datadir, model_name)
        confirm_msg = f"\nYou are about to download the LLM {model_name} to the {datadir} folder. Are you sure?"
        if os.path.isfile(filename):
            confirm_msg = f"There is already a file {model_name} in {datadir}.\n Do you want to still download it?"

        shall = True
        if confirm:
            shall = input("%s (y/N) " % confirm_msg).lower() == "y"
        if shall:
            U.download(model_url, filename, verify=ssl_verify)
        else:
            warnings.warn(
                f'{model_name} was not downloaded because "Y" was not selected.'
            )
        return

    def load_ingester(self):
        """
        Get `Ingester` instance.
        You can access the `langchain_chroma.Chroma` instance with `load_ingester().get_db()`.
        """
        if not self.ingester:
            from onprem.ingest import Ingester

            self.ingester = Ingester(
                embedding_model_name=self.embedding_model_name,
                embedding_model_kwargs=self.embedding_model_kwargs,
                embedding_encode_kwargs=self.embedding_encode_kwargs,
                persist_directory=self.vectordb_path,
            )
        return self.ingester

    def load_vectordb(self):
        """
        Get Chroma db instance
        """
        ingester = self.load_ingester()
        db = ingester.get_db()
        if not db:
            raise ValueError(
                "A vector database has not yet been created. Please call the LLM.ingest method."
            )
        return db

    def ingest(
        self,
        source_directory: str, # path to folder containing documents
        chunk_size: int = 500, # text is split to this many characters by `langchain.text_splitter.RecursiveCharacterTextSplitter`
        chunk_overlap: int = 50, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
        ignore_fn:Optional[Callable] = None, # callable that accepts the file path and returns True for ignored files
        pdf_use_unstructured:bool=False, # If True, use unstructured for PDF extraction
        **kwargs, # Extra kwargs fed to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True
    ):
        """
        Ingests all documents in `source_folder` into vector database.
        Previously-ingested documents are ignored.
        Extra kwargs fed directly to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True.
        """
        ingester = self.load_ingester()
        return ingester.ingest(
            source_directory,
            chunk_size=chunk_size, chunk_overlap=chunk_overlap, ignore_fn=ignore_fn,
            pdf_use_unstructured=pdf_use_unstructured,
            **kwargs
        )

    def check_model(self):
        """
        Returns the path to the model
        """
        if not self.is_llamacpp():
            return None
        datadir = self.model_download_path
        model_path = os.path.join(datadir, self.model_name)
        if not os.path.isfile(model_path):
            raise ValueError(
                f"The LLM model {self.model_name} does not appear to have been downloaded. "
                + "Execute the download_model() method to download it."
            )
        return model_path

    def load_llm(self):
        """
        Loads the LLM from the model path.
        """

        if not self.llm and self.is_openai_model():
            self.llm = ChatOpenAI(model_name=self.model_name, 
                                  callbacks=self.callbacks, 
                                  streaming=not self.mute_stream,
                                  max_tokens=self.max_tokens,
                                  **self.extra_kwargs)
        elif not self.llm and self.is_azure():
            self.llm = AzureChatOpenAI(azure_deployment=self.model_name, 
                                  callbacks=self.callbacks, 
                                  streaming=not self.mute_stream,
                                  max_tokens=self.max_tokens,
                                  **self.extra_kwargs)

        elif not self.llm and self.is_local_api():
            self.llm = ChatOpenAI(base_url=self.model_url,
                                  #model_name=self.model_name, 
                                  callbacks=self.callbacks, 
                                  streaming=not self.mute_stream,
                                  max_tokens=self.max_tokens,
                                  **self.extra_kwargs)
        elif not self.llm and self.is_hf():
            from onprem.hf import LLM
            from transformers import TextStreamer, AutoTokenizer
            from langchain_huggingface import ChatHuggingFace, HuggingFacePipeline
            tokenizer = AutoTokenizer.from_pretrained(self.model_id)
            streamer = TextStreamer(tokenizer)
            llmpipe = LLM(self.model_id,
                          streamer=streamer,
                          max_new_tokens = self.max_tokens,
                          do_sample=True if\
                                    self.extra_kwargs.get('temperature', 0.8)>0.0 else False ,
                          **self.extra_kwargs)
            hfpipe = HuggingFacePipeline(pipeline=llmpipe.generator.llm.pipeline)
            self.llm = ChatHuggingFace(llm=hfpipe)

        elif not self.llm:
            model_path = self.check_model()
            if self.check_model_download and os.path.getsize(model_path) < MIN_MODEL_SIZE:
                raise ValueError(f'The model file ({model_path} is less than {MIN_MODEL_SIZE} bytes. ' +\
                                 'It may not have been fully downloaded. Please delete the file and start again. ')
            self.llm = LlamaCpp(
                model_path=model_path,
                max_tokens=self.max_tokens,
                n_batch=self.n_batch,
                callbacks=self.callbacks,
                verbose=self.verbose,
                n_gpu_layers=self.n_gpu_layers,
                n_ctx=self.n_ctx,
                #offload_kqv = self.offload_kqv,
                **self.extra_kwargs,
            )

        return self.llm


    def prompt(self,
               prompt:Union[str, List[Dict]],
               image_path_or_url:Optional[str] = None,
               prompt_template: Optional[str] = None, stop:list=[], **kwargs):
        """
        Send prompt to LLM to generate a response.
        Extra keyword arguments are sent directly to the model invocation.

        **Args:**

        - *prompt*: The prompt to supply to the model.
                    Either a string or OpenAI-style list of dictionaries
                    representing messages (e.g., "human", "system").
        - *image_path_or_url*: Path or URL to an image file
        - *prompt_template*: Optional prompt template (must have a variable named "prompt").
                             This value will override any `prompt_template` value supplied 
                             to `LLM` constructor.
        - *stop*: a list of strings to stop generation when encountered. 
                  This value will override the `stop` parameter supplied to `LLM` constructor.

        """
        if isinstance(prompt, list): # list of dictionaries representing messages
            try:
                res = self.llm.invoke(prompt, stop=stop, **kwargs)
            except Exception as e: # stop param fails with GPT-4o vision prompts
                res = self.llm.invoke(prompt, **kwargs)
        else:
            from langchain_core.messages import HumanMessage
            import base64
            llm = self.load_llm()
            prompt_template = self.prompt_template if prompt_template is None else prompt_template
            if prompt_template:
                prompt = prompt_template.format(**{"prompt": prompt})
            stop = stop if stop else self.stop
            if image_path_or_url:
                if not image_path_or_url.startswith('http'):
                    with open(image_path_or_url, "rb") as f:
                        image_data = base64.b64encode(f.read()).decode('utf-8')
                    image_path_or_url = f"data:image/jpeg;base64,{image_data}"

                message = HumanMessage(
                    content=[
                        {"type": "text", "text": prompt},
                        {
                            "type": "image_url",
                            "image_url": {"url": image_path_or_url},
                        },
                    ],
                )
                prompt = [message]
                res = llm.invoke(prompt, **kwargs) # including stop causes errors in gpt-4o
            else:
                res = llm.invoke(prompt, stop=stop, **kwargs)
        return res.content if self.is_openai_model() or self.is_hf() else res



    def load_qa(self, prompt_template: str = DEFAULT_QA_PROMPT):
        """
        Prepares and loads the `langchain.chains.RetrievalQA` object

        **Args:**

        - *prompt_template*: A string representing the prompt with variables "context" and "question"
        """
        if self.qa is None:
            db = self.load_vectordb()
            retriever = db.as_retriever(
                search_type="similarity_score_threshold",
                search_kwargs={
                    "k": self.rag_num_source_docs,
                    "score_threshold": self.rag_score_threshold,
                },
            )
            llm = self.load_llm()
            PROMPT = PromptTemplate(
                template=prompt_template, input_variables=["context", "question"]
            )
            self.qa = RetrievalQA.from_chain_type(
                llm=llm,
                chain_type="stuff",
                retriever=retriever,
                return_source_documents=True,
                chain_type_kwargs={"prompt": PROMPT},
            )
        return self.qa

    def load_chatqa(self):
        """
        Prepares and loads a `langchain.chains.ConversationalRetrievalChain` instance
        """
        if self.chatqa is None:
            db = self.load_vectordb()
            retriever = db.as_retriever(
                search_type="similarity_score_threshold",  # see note in constructor
                search_kwargs={
                    "k": self.rag_num_source_docs,
                    "score_threshold": self.rag_score_threshold,
                },
            )
            llm = self.load_llm()
            memory = AnswerConversationBufferMemory(
                memory_key="chat_history", return_messages=True
            )
            self.chatqa = ConversationalRetrievalChain.from_llm(
                llm, retriever, memory=memory, return_source_documents=True
            )
        return self.chatqa

    def ask(self, question: str, qa_template=DEFAULT_QA_PROMPT, prompt_template=None, **kwargs):
        """
        Answer a question based on source documents fed to the `ingest` method.
        Extra keyword arguments are sent directly to the model invocation.

        **Args:**

        - *question*: a question you want to ask
        - *qa_template*: A string representing the prompt with variables "context" and "question"
        - *prompt_template*: the model-specific template in which everything (including QA template) should be wrapped.
                            Should have a single variable "{prompt}". Overrides the `prompt_template` parameter supplied to 
                            `LLM` constructor.

        **Returns:**

        - A dictionary with keys: `answer`, `source_documents`, `question`
        """
        prompt_template = self.prompt_template if prompt_template is None else prompt_template
        prompt_template = qa_template if prompt_template is None else prompt_template.format(**{'prompt': qa_template})
        qa = self.load_qa(prompt_template=prompt_template)
        res = qa.invoke(question, **kwargs)
        res["question"] = res["query"]
        del res["query"]
        res["answer"] = res["result"]
        del res["result"]
        return res

    def chat(self, question: str, **kwargs):
        """
        Chat with documents fed to the `ingest` method.
        Unlike `LLM.ask`, `LLM.chat` includes conversational memory.
        Extra keyword arguments are sent directly to the model invocation.

        **Args:**

        - *question*: a question you want to ask

        **Returns:**

        - A dictionary with keys: `answer`, `source_documents`, `question`, `chat_history`
        """
        chatqa = self.load_chatqa()
        res = chatqa.invoke(question, **kwargs)
        return res



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/onprem',
                'doc_host': 'https://amaiya.github.io',
                'git_url': 'https://github.com/amaiya/onprem',
                'lib_path': 'onprem'},
  'syms': { 'onprem.console': {},
            'onprem.core': { 'onprem.core.AnswerConversationBufferMemory': ('core.html#answerconversationbuffermemory', 'onprem/core.py'),
                             'onprem.core.AnswerConversationBufferMemory.save_context': ( 'core.html#answerconversationbuffermemory.save_context',
                                                                                          'onprem/core.py'),
                             'onprem.core.LLM': ('core.html#llm', 'onprem/core.py'),
                             'onprem.core.LLM.__init__': ('core.html#llm.__init__', 'onprem/core.py'),
                             'onprem.core.LLM.ask': ('core.html#llm.ask', 'onprem/core.py'),
                             'onprem.core.LLM.chat': ('core.html#llm.chat', 'onprem/core.py'),
                             'onprem.core.LLM.check_model': ('core.html#llm.check_model', 'onprem/core.py'),
                             'onprem.core.LLM.download_model': ('core.html#llm.download_model', 'onprem/core.py'),
                             'onprem.core.LLM.ingest': ('core.html#llm.ingest', 'onprem/core.py'),
                             'onprem.core.LLM.is_azure': ('core.html#llm.is_azure', 'onprem/core.py'),
                             'onprem.core.LLM.is_hf': ('core.html#llm.is_hf', 'onprem/core.py'),
                             'onprem.core.LLM.is_llamacpp': ('core.html#llm.is_llamacpp', 'onprem/core.py'),
                             'onprem.core.LLM.is_local': ('core.html#llm.is_local', 'onprem/core.py'),
                             'onprem.core.LLM.is_local_api': ('core.html#llm.is_local_api', 'onprem/core.py'),
                             'onprem.core.LLM.is_openai_model': ('core.html#llm.is_openai_model', 'onprem/core.py'),
                             'onprem.core.LLM.load_chatqa': ('core.html#llm.load_chatqa', 'onprem/core.py'),
                             'onprem.core.LLM.load_ingester': ('core.html#llm.load_ingester', 'onprem/core.py'),
                             'onprem.core.LLM.load_llm': ('core.html#llm.load_llm', 'onprem/core.py'),
                             'onprem.core.LLM.load_qa': ('core.html#llm.load_qa', 'onprem/core.py'),
                             'onprem.core.LLM.load_vectordb': ('core.html#llm.load_vectordb', 'onprem/core.py'),
                             'onprem.core.LLM.prompt': ('core.html#llm.prompt', 'onprem/core.py'),
                             'onprem.core.LLM.update_max_tokens': ('core.html#llm.update_max_tokens', 'onprem/core.py'),
                             'onprem.core.LLM.update_stop': ('core.html#llm.update_stop', 'onprem/core.py')},
            'onprem.guider': { 'onprem.guider.Guider': ('guider.html#guider', 'onprem/guider.py'),
                               'onprem.guider.Guider.__init__': ('guider.html#guider.__init__', 'onprem/guider.py'),
                               'onprem.guider.Guider.prompt': ('guider.html#guider.prompt', 'onprem/guider.py')},
            'onprem.hf.hfmodel': { 'onprem.hf.hfmodel.HFModel': ('hf.hfmodels.html#hfmodel', 'onprem/hf/hfmodel.py'),
                                   'onprem.hf.hfmodel.HFModel.__init__': ('hf.hfmodels.html#hfmodel.__init__', 'onprem/hf/hfmodel.py'),
                                   'onprem.hf.hfmodel.HFModel.prepare': ('hf.hfmodels.html#hfmodel.prepare', 'onprem/hf/hfmodel.py'),
                                   'onprem.hf.hfmodel.HFModel.tokenize': ('hf.hfmodels.html#hfmodel.tokenize', 'onprem/hf/hfmodel.py')},
            'onprem.hf.hfpipeline': { 'onprem.hf.hfpipeline.HFPipeline': ('hf.hfpipeline.html#hfpipeline', 'onprem/hf/hfpipeline.py'),
                                      'onprem.hf.hfpipeline.HFPipeline.__init__': ( 'hf.hfpipeline.html#hfpipeline.__init__',
                                                                                    'onprem/hf/hfpipeline.py'),
                                      'onprem.hf.hfpipeline.HFPipeline.maxlength': ( 'hf.hfpipeline.html#hfpipeline.maxlength',
                                                                                     'onprem/hf/hfpipeline.py'),
                                      'onprem.hf.hfpipeline.HFPipeline.parseargs': ( 'hf.hfpipeline.html#hfpipeline.parseargs',
                                                                                     'onprem/hf/hfpipeline.py')},
            'onprem.hf.llm.factory': { 'onprem.hf.llm.factory.GenerationFactory': ( 'hf.llm.factory.html#generationfactory',
                                                                                    'onprem/hf/llm/factory.py'),
                                       'onprem.hf.llm.factory.GenerationFactory.create': ( 'hf.llm.factory.html#generationfactory.create',
                                                                                           'onprem/hf/llm/factory.py'),
                                       'onprem.hf.llm.factory.GenerationFactory.method': ( 'hf.llm.factory.html#generationfactory.method',
                                                                                           'onprem/hf/llm/factory.py'),
                                       'onprem.hf.llm.factory.GenerationFactory.resolve': ( 'hf.llm.factory.html#generationfactory.resolve',
                                                                                            'onprem/hf/llm/factory.py')},
            'onprem.hf.llm.generation': { 'onprem.hf.llm.generation.Generation': ( 'hf.llm.generation.html#generation',
                                                                                   'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.__call__': ( 'hf.llm.generation.html#generation.__call__',
                                                                                            'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.__init__': ( 'hf.llm.generation.html#generation.__init__',
                                                                                            'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.clean': ( 'hf.llm.generation.html#generation.clean',
                                                                                         'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.execute': ( 'hf.llm.generation.html#generation.execute',
                                                                                           'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.response': ( 'hf.llm.generation.html#generation.response',
                                                                                            'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.stream': ( 'hf.llm.generation.html#generation.stream',
                                                                                          'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.TemplateFormatter': ( 'hf.llm.generation.html#templateformatter',
                                                                                          'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.TemplateFormatter.check_unused_args': ( 'hf.llm.generation.html#templateformatter.check_unused_args',
                                                                                                            'onprem/hf/llm/generation.py')},
            'onprem.hf.llm.huggingface': { 'onprem.hf.llm.huggingface.Generator': ( 'hf.llm.huggingface.html#generator',
                                                                                    'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.Generator.__init__': ( 'hf.llm.huggingface.html#generator.__init__',
                                                                                             'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFGeneration': ( 'hf.llm.huggingface.html#hfgeneration',
                                                                                       'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFGeneration.__init__': ( 'hf.llm.huggingface.html#hfgeneration.__init__',
                                                                                                'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFGeneration.stream': ( 'hf.llm.huggingface.html#hfgeneration.stream',
                                                                                              'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM': ( 'hf.llm.huggingface.html#hfllm',
                                                                                'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.__call__': ( 'hf.llm.huggingface.html#hfllm.__call__',
                                                                                         'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.__init__': ( 'hf.llm.huggingface.html#hfllm.__init__',
                                                                                         'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.extract': ( 'hf.llm.huggingface.html#hfllm.extract',
                                                                                        'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.parameters': ( 'hf.llm.huggingface.html#hfllm.parameters',
                                                                                           'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.task': ( 'hf.llm.huggingface.html#hfllm.task',
                                                                                     'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.Sequences': ( 'hf.llm.huggingface.html#sequences',
                                                                                    'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.Sequences.__init__': ( 'hf.llm.huggingface.html#sequences.__init__',
                                                                                             'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse': ( 'hf.llm.huggingface.html#streamingresponse',
                                                                                            'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse.__call__': ( 'hf.llm.huggingface.html#streamingresponse.__call__',
                                                                                                     'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse.__init__': ( 'hf.llm.huggingface.html#streamingresponse.__init__',
                                                                                                     'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse.__iter__': ( 'hf.llm.huggingface.html#streamingresponse.__iter__',
                                                                                                     'onprem/hf/llm/huggingface.py')},
            'onprem.hf.llm.llm': { 'onprem.hf.llm.llm.LLM': ('hf.llm.llm.html#llm', 'onprem/hf/llm/llm.py'),
                                   'onprem.hf.llm.llm.LLM.__call__': ('hf.llm.llm.html#llm.__call__', 'onprem/hf/llm/llm.py'),
                                   'onprem.hf.llm.llm.LLM.__init__': ('hf.llm.llm.html#llm.__init__', 'onprem/hf/llm/llm.py')},
            'onprem.hf.models': { 'onprem.hf.models.Models': ('hf.models.html#models', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.checklength': ('hf.models.html#models.checklength', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.device': ('hf.models.html#models.device', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.deviceid': ('hf.models.html#models.deviceid', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.finddevice': ('hf.models.html#models.finddevice', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.hasaccelerator': ('hf.models.html#models.hasaccelerator', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.hasmpsdevice': ('hf.models.html#models.hasmpsdevice', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.load': ('hf.models.html#models.load', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.maxlength': ('hf.models.html#models.maxlength', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.reference': ('hf.models.html#models.reference', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.task': ('hf.models.html#models.task', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.tokenizer': ('hf.models.html#models.tokenizer', 'onprem/hf/models.py')},
            'onprem.hf.onnx': { 'onprem.hf.onnx.OnnxConfig': ('hf.onnx.html#onnxconfig', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel': ('hf.onnx.html#onnxmodel', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.__init__': ('hf.onnx.html#onnxmodel.__init__', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.device': ('hf.onnx.html#onnxmodel.device', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.forward': ('hf.onnx.html#onnxmodel.forward', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.parse': ('hf.onnx.html#onnxmodel.parse', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.providers': ('hf.onnx.html#onnxmodel.providers', 'onprem/hf/onnx.py')},
            'onprem.hf.registry': { 'onprem.hf.registry.Registry': ('hf.registry.html#registry', 'onprem/hf/registry.py'),
                                    'onprem.hf.registry.Registry.register': ( 'hf.registry.html#registry.register',
                                                                              'onprem/hf/registry.py')},
            'onprem.hf.resolver': { 'onprem.hf.resolver.Resolver': ('hf.resolver.html#resolver', 'onprem/hf/resolver.py'),
                                    'onprem.hf.resolver.Resolver.__call__': ( 'hf.resolver.html#resolver.__call__',
                                                                              'onprem/hf/resolver.py')},
            'onprem.hf.tensors': { 'onprem.hf.tensors.Pipeline': ('hf.tensors.html#pipeline', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Pipeline.batch': ('hf.tensors.html#pipeline.batch', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors': ('hf.tensors.html#tensors', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.context': ('hf.tensors.html#tensors.context', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.quantize': ('hf.tensors.html#tensors.quantize', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.tensor': ('hf.tensors.html#tensors.tensor', 'onprem/hf/tensors.py')},
            'onprem.ingest': { 'onprem.ingest.Ingester': ('ingest.html#ingester', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.__init__': ('ingest.html#ingester.__init__', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_db': ('ingest.html#ingester.get_db', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_embedding_model': ( 'ingest.html#ingester.get_embedding_model',
                                                                               'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_ingested_files': ('ingest.html#ingester.get_ingested_files', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.ingest': ('ingest.html#ingester.ingest', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.store_documents': ('ingest.html#ingester.store_documents', 'onprem/ingest.py'),
                               'onprem.ingest.MyElmLoader': ('ingest.html#myelmloader', 'onprem/ingest.py'),
                               'onprem.ingest.MyElmLoader.load': ('ingest.html#myelmloader.load', 'onprem/ingest.py'),
                               'onprem.ingest.MyPDFLoader': ('ingest.html#mypdfloader', 'onprem/ingest.py'),
                               'onprem.ingest.MyPDFLoader.load': ('ingest.html#mypdfloader.load', 'onprem/ingest.py'),
                               'onprem.ingest.batchify_chunks': ('ingest.html#batchify_chunks', 'onprem/ingest.py'),
                               'onprem.ingest.does_vectorstore_exist': ('ingest.html#does_vectorstore_exist', 'onprem/ingest.py'),
                               'onprem.ingest.extract_files': ('ingest.html#extract_files', 'onprem/ingest.py'),
                               'onprem.ingest.load_documents': ('ingest.html#load_documents', 'onprem/ingest.py'),
                               'onprem.ingest.load_single_document': ('ingest.html#load_single_document', 'onprem/ingest.py'),
                               'onprem.ingest.process_documents': ('ingest.html#process_documents', 'onprem/ingest.py')},
            'onprem.pipelines.classifier': { 'onprem.pipelines.classifier.ClassifierBase': ( 'pipelines.classifier.html#classifierbase',
                                                                                             'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.arrays2dataset': ( 'pipelines.classifier.html#classifierbase.arrays2dataset',
                                                                                                            'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.dataset2arrays': ( 'pipelines.classifier.html#classifierbase.dataset2arrays',
                                                                                                            'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.evaluate': ( 'pipelines.classifier.html#classifierbase.evaluate',
                                                                                                      'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.explain': ( 'pipelines.classifier.html#classifierbase.explain',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_labels': ( 'pipelines.classifier.html#classifierbase.get_labels',
                                                                                                        'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_trainer': ( 'pipelines.classifier.html#classifierbase.get_trainer',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.sample_examples': ( 'pipelines.classifier.html#classifierbase.sample_examples',
                                                                                                             'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.save': ( 'pipelines.classifier.html#classifierbase.save',
                                                                                                  'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.train': ( 'pipelines.classifier.html#classifierbase.train',
                                                                                                   'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier': ( 'pipelines.classifier.html#fewshotclassifier',
                                                                                                'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.__init__': ( 'pipelines.classifier.html#fewshotclassifier.__init__',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.save': ( 'pipelines.classifier.html#fewshotclassifier.save',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.train': ( 'pipelines.classifier.html#fewshotclassifier.train',
                                                                                                      'onprem/pipelines/classifier.py')},
            'onprem.pipelines.extractor': { 'onprem.pipelines.extractor.Extractor': ( 'pipelines.extractor.html#extractor',
                                                                                      'onprem/pipelines/extractor.py'),
                                            'onprem.pipelines.extractor.Extractor.__init__': ( 'pipelines.extractor.html#extractor.__init__',
                                                                                               'onprem/pipelines/extractor.py'),
                                            'onprem.pipelines.extractor.Extractor.apply': ( 'pipelines.extractor.html#extractor.apply',
                                                                                            'onprem/pipelines/extractor.py')},
            'onprem.pipelines.summarizer': { 'onprem.pipelines.summarizer.Summarizer': ( 'pipelines.summarizer.html#summarizer',
                                                                                         'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.__init__': ( 'pipelines.summarizer.html#summarizer.__init__',
                                                                                                  'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer._map_reduce': ( 'pipelines.summarizer.html#summarizer._map_reduce',
                                                                                                     'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer._refine': ( 'pipelines.summarizer.html#summarizer._refine',
                                                                                                 'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.summarize': ( 'pipelines.summarizer.html#summarizer.summarize',
                                                                                                   'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.summarize_by_concept': ( 'pipelines.summarizer.html#summarizer.summarize_by_concept',
                                                                                                              'onprem/pipelines/summarizer.py')},
            'onprem.utils': { 'onprem.utils.df_to_md': ('utils.html#df_to_md', 'onprem/utils.py'),
                              'onprem.utils.download': ('utils.html#download', 'onprem/utils.py'),
                              'onprem.utils.get_datadir': ('utils.html#get_datadir', 'onprem/utils.py'),
                              'onprem.utils.html_to_df': ('utils.html#html_to_df', 'onprem/utils.py'),
                              'onprem.utils.md_to_df': ('utils.html#md_to_df', 'onprem/utils.py'),
                              'onprem.utils.segment': ('utils.html#segment', 'onprem/utils.py'),
                              'onprem.utils.split_list': ('utils.html#split_list', 'onprem/utils.py')},
            'onprem.webapp': {}}}



"""functionality for document ingestion into a vector database for question-answering"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_ingest.ipynb.

# %% auto 0
__all__ = ['logger', 'DEFAULT_CHUNK_SIZE', 'DEFAULT_CHUNK_OVERLAP', 'COLLECTION_NAME', 'CHROMA_MAX', 'LOADER_MAPPING',
           'DEFAULT_DB', 'MyElmLoader', 'MyPDFLoader', 'extract_files', 'load_single_document', 'load_documents',
           'process_documents', 'does_vectorstore_exist', 'batchify_chunks', 'Ingester']

# %% ../nbs/01_ingest.ipynb 3
from .utils import get_datadir
import os
import os.path
import glob
from typing import List, Optional, Callable
from multiprocessing import Pool
import functools
from tqdm import tqdm
import warnings

from langchain_core.documents import Document
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import (
    CSVLoader,
    EverNoteLoader,
    TextLoader,
    PyMuPDFLoader,
    UnstructuredPDFLoader,
    UnstructuredEmailLoader,
    UnstructuredEPubLoader,
    UnstructuredHTMLLoader,
    UnstructuredMarkdownLoader,
    UnstructuredODTLoader,
    UnstructuredPowerPointLoader,
    UnstructuredWordDocumentLoader,
)
from langchain_chroma import Chroma
from langchain_huggingface import HuggingFaceEmbeddings
import chromadb
from chromadb.config import Settings
from . import utils as U

import logging
import sys

logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
logger = logging.getLogger('OnPrem.LLM-ingest')

DEFAULT_CHUNK_SIZE = 500
DEFAULT_CHUNK_OVERLAP = 50
COLLECTION_NAME = "onprem_chroma"
CHROMA_MAX = 41000

# %% ../nbs/01_ingest.ipynb 4
class MyElmLoader(UnstructuredEmailLoader):
    """Wrapper to fallback to text/plain when default does not work"""

    def load(self) -> List[Document]:
        """Wrapper adding fallback for elm without html"""
        try:
            try:
                doc = UnstructuredEmailLoader.load(self)
            except ValueError as e:
                if "text/html content not found in email" in str(e):
                    # Try plain text
                    self.unstructured_kwargs["content_source"] = "text/plain"
                    doc = UnstructuredEmailLoader.load(self)
                else:
                    raise
        except Exception as e:
            raise Exception(f'{self.file_path} : {e}')

        return doc

class MyPDFLoader(UnstructuredPDFLoader):
    """Custom Unstructured-based PDF Loader"""

    def load(self) -> List[Document]:
        """Wrapper adding fallback for elm without html"""
        #if self.unstructured_kwargs.get('mode', None) != 'elements':
            #raise ValueError('MyPDFLoader needs to be called with mode="elements"')
        try:
            docs = UnstructuredPDFLoader.load(self)
            if not docs:
                raise Exception('Document had no content. ')
            page_content = '\n'.join([d.metadata.get('text_as_html', d.page_content)
                           if d.metadata.get('category', None) == 'Table' else d.page_content for d in docs])
            source = docs[0].metadata['source']
            doc = Document(page_content=page_content, metadata={'source':source})
            return [doc]
        except Exception as e:
            # Add file_path to exception message
            raise Exception(f'{self.file_path} : {e}')


# %% ../nbs/01_ingest.ipynb 5
# Map file extensions to document loaders and their arguments
LOADER_MAPPING = {
    ".csv": (CSVLoader, {}),
    ".doc": (UnstructuredWordDocumentLoader, {}),
    ".docx": (UnstructuredWordDocumentLoader, {}),
    ".enex": (EverNoteLoader, {}),
    ".eml": (MyElmLoader, {}),
    ".epub": (UnstructuredEPubLoader, {}),
    ".html": (UnstructuredHTMLLoader, {}),
    ".md": (UnstructuredMarkdownLoader, {}),
    ".odt": (UnstructuredODTLoader, {}),
    ".pdf": (PyMuPDFLoader, {}),
    #".pdfOCR": (UnstructuredPDFLoader, {"strategy":"hi_res"}),
    ".pdfOCR": (MyPDFLoader, {"infer_table_structure":False, "mode":"elements", "strategy":"hi_res"}),
    ".ppt": (UnstructuredPowerPointLoader, {}),
    ".pptx": (UnstructuredPowerPointLoader, {}),
    ".txt": (TextLoader, {"encoding": "utf8"}),
    # Add more mappings for other file extensions and loaders as needed
}


def extract_files(source_dir:str):
    """
    Extract files of supported file types from folder.
    """
    source_dir = os.path.abspath(source_dir)
    all_files = []
    for ext in LOADER_MAPPING:
        all_files.extend(
            glob.glob(os.path.join(source_dir, f"**/*{ext.lower()}"), recursive=True)
        )
        all_files.extend(
            glob.glob(os.path.join(source_dir, f"**/*{ext.upper()}"), recursive=True)
        )
    return all_files


def load_single_document(file_path: str, # path to file
                         pdf_use_unstructured:bool=False, # use unstructured for PDF extraction if True
                         **kwargs,
                         ) -> List[Document]:
    """
    Load a single document. Will attempt to OCR PDFs, if necessary.
    Extra kwargs fed to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True.
    """
    ext = "." + file_path.rsplit(".", 1)[-1].lower()
    if ext in LOADER_MAPPING:
        try:
            if 'strategy' in kwargs or 'infer_table_structure' in kwargs:
                # auto-set pdf_use_unstructured if above key config vars are supplied
                pdf_use_unstructured = True

            loader_class, loader_args = LOADER_MAPPING[ext+f'{"OCR" if pdf_use_unstructured and ext==".pdf" else ""}']
            loader_args = loader_args.copy() # copy so any supplied kwargs do not persist across calls
            if pdf_use_unstructured and ext=='.pdf':
                loader_args.update(kwargs)
            loader = loader_class(file_path, **loader_args)
            if ext == '.pdf' and not pdf_use_unstructured:
                with warnings.catch_warnings():
                    warnings.simplefilter(action='ignore', category=UserWarning)
                    docs = loader.load()
                if not docs or len('\n'.join([d.page_content.strip() for d in docs]).strip()) < 32:
                    loader_class, loader_args = LOADER_MAPPING[ext+'OCR']
                    loader = loader_class(file_path, **loader_args)
                    docs = loader.load()
                    for doc in docs:
                        doc.metadata = dict(doc.metadata, ocr=True)
                return docs
            else:
                return loader.load()

        except Exception as e:
            logger.warning(f'\nSkipping {file_path} due to error: {str(e)}')
    else:
        logger.warning(f"\nSkipping {file_path} due to unsupported file extension: '{ext}'")


def load_documents(source_dir: str, # path to folder containing documents
                   ignored_files: List[str] = [], # list of filepaths to ignore
                   ignore_fn:Optional[Callable] = None, # callable that accepts file path and returns True for ignored files
                   pdf_use_unstructured:bool=False, # If True, use unstructured for PDF extraction
                   **kwargs
) -> List[Document]:
    """
    Loads all documents from the source documents directory, ignoring specified files.
    Extra kwargs fed to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True.
    """
    all_files = extract_files(source_dir)

    filtered_files = [
        file_path for file_path in all_files if file_path not in ignored_files and not os.path.basename(file_path).startswith('~$')
         and (ignore_fn is None or not ignore_fn(file_path))
    ]

    with Pool(processes=os.cpu_count()) as pool:
        results = []
        with tqdm(
            total=len(filtered_files), desc="Loading new documents", ncols=80
        ) as pbar:
            for i, docs in enumerate(
                pool.imap_unordered(functools.partial(load_single_document,
                                                      pdf_use_unstructured=pdf_use_unstructured, **kwargs),
                                    filtered_files)
            ):
                if docs is not None:
                    results.extend(docs)
                pbar.update()

    return results


def process_documents(
    source_directory: str, # path to folder containing document store
    chunk_size: int = DEFAULT_CHUNK_SIZE, # text is split to this many characters by `langchain.text_splitter.RecursiveCharacterTextSplitter`
    chunk_overlap: int = DEFAULT_CHUNK_OVERLAP, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
    ignored_files: List[str] = [], # list of files to ignore
    ignore_fn:Optional[Callable] = None, # Callable that accepts the file path (including file name) as input and ignores if returns True
    pdf_use_unstructured:bool=False, # If True, use unstructured for PDF extraction
    **kwargs


) -> List[Document]:
    """
    Load documents and split in chunks.
    Extra kwargs fed to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True
    """
    print(f"Loading documents from {source_directory}")
    documents = load_documents(source_directory,
                              ignored_files, ignore_fn=ignore_fn,
                              pdf_use_unstructured=pdf_use_unstructured, **kwargs)
    if not documents:
        print("No new documents to load")
        return
    print(f"Loaded {len(documents)} new documents from {source_directory}")
    text_splitter = RecursiveCharacterTextSplitter(
        chunk_size=chunk_size, chunk_overlap=chunk_overlap
    )
    texts = text_splitter.split_documents(documents)
    print(f"Split into {len(texts)} chunks of text (max. {chunk_size} chars each)")
    return texts


def does_vectorstore_exist(db) -> bool:
    """
    Checks if vectorstore exists
    """
    if not db.get()["documents"]:
        return False
    return True


def batchify_chunks(texts):
    """
    split texts into batches specifically for Chroma
    """
    split_docs_chunked = U.split_list(texts, CHROMA_MAX)
    total_chunks = sum(1 for _ in U.split_list(texts, CHROMA_MAX))
    return split_docs_chunked, total_chunks



os.environ["TOKENIZERS_PARALLELISM"] = "0"
DEFAULT_DB = "vectordb"


class Ingester:
    def __init__(
        self,
        embedding_model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
        embedding_model_kwargs: dict = {"device": "cpu"},
        embedding_encode_kwargs: dict = {"normalize_embeddings": False},
        persist_directory: Optional[str] = None,
    ):
        """
        Ingests all documents in `source_folder` (previously-ingested documents are ignored)

        **Args**:

          - *embedding_model*: name of sentence-transformers model
          - *embedding_model_kwargs*: arguments to embedding model (e.g., `{device':'cpu'}`)
          - *embedding_encode_kwargs*: arguments to encode method of
                                       embedding model (e.g., `{'normalize_embeddings': False}`).
          - *persist_directory*: Path to vector database (created if it doesn't exist).
                                 Default is `onprem_data/vectordb` in user's home directory.


        **Returns**: `None`
        """
        self.persist_directory = persist_directory or os.path.join(
            get_datadir(), DEFAULT_DB
        )
        self.embeddings = HuggingFaceEmbeddings(
            model_name=embedding_model_name,
            model_kwargs=embedding_model_kwargs,
            encode_kwargs=embedding_encode_kwargs,
        )
        self.chroma_settings = Settings(
            persist_directory=self.persist_directory, anonymized_telemetry=False
        )
        self.chroma_client = chromadb.PersistentClient(
            settings=self.chroma_settings, path=self.persist_directory
        )
        return

    def get_db(self):
        """
        Returns an instance to the `langchain_chroma.Chroma` instance
        """
        db = Chroma(
            persist_directory=self.persist_directory,
            embedding_function=self.embeddings,
            client_settings=self.chroma_settings,
            client=self.chroma_client,
            collection_metadata={"hnsw:space": "cosine"},
            collection_name=COLLECTION_NAME,
        )
        return db if does_vectorstore_exist(db) else None

    def get_embedding_model(self):
        """
        Returns an instance to the `langchain_huggingface.HuggingFaceEmbeddings` instance
        """
        return self.embeddings


    def get_ingested_files(self):
        """
        Returns a list of files previously added to vector database (typically via `LLM.ingest`)
        """
        return set([d['source'] for d in self.get_db().get()['metadatas']])


    def store_documents(self, documents):
        """
        Stores instances of `langchain_core.documents.base.Document` in vectordb
        """
        if not documents:
            return
        db = self.get_db()
        if db:
            print("Creating embeddings. May take some minutes...")
            chunk_batches, total_chunks = batchify_chunks(documents)
            for lst in tqdm(chunk_batches, total=total_chunks):
                db.add_documents(lst)
        else:
            chunk_batches, total_chunks = batchify_chunks(documents)
            print("Creating embeddings. May take some minutes...")
            db = None

            for lst in tqdm(chunk_batches, total=total_chunks):
                if not db:
                    db = Chroma.from_documents(
                        lst,
                        self.embeddings,
                        persist_directory=self.persist_directory,
                        client_settings=self.chroma_settings,
                        client=self.chroma_client,
                        collection_metadata={"hnsw:space": "cosine"},
                        collection_name=COLLECTION_NAME,
                    )
                else:
                    db.add_documents(lst)
        return


    def ingest(
        self,
        source_directory: str, # path to folder containing document store
        chunk_size: int = DEFAULT_CHUNK_SIZE, # text is split to this many characters by [langchain.text_splitter.RecursiveCharacterTextSplitter](https://api.python.langchain.com/en/latest/character/langchain_text_splitters.character.RecursiveCharacterTextSplitter.html)
        chunk_overlap: int = DEFAULT_CHUNK_OVERLAP, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
        ignore_fn:Optional[Callable] = None, # Optional function that accepts the file path (including file name) as input and returns `True` if file path should not be ingested.
        pdf_use_unstructured:bool=False, # If True, use unstructured for PDF extraction
        **kwargs
    ) -> None:
        """
        Ingests all documents in `source_directory` (previously-ingested documents are
        ignored). When retrieved, the
        [Document](https://api.python.langchain.com/en/latest/documents/langchain_core.documents.base.Document.html)
        objects will each have a `metadata` dict with the absolute path to the file
        in `metadata["source"]`.
        Extra kwargs fed to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True
        """

        if not os.path.exists(source_directory):
            raise ValueError("The source_directory does not exist.")
        elif os.path.isfile(source_directory):
            raise ValueError(
                "The source_directory argument must be a folder, not a file."
            )
        texts = None
        db = self.get_db()
        if db:
            # Update and store locally vectorstore
            print(f"Appending to existing vectorstore at {self.persist_directory}")
            collection = db.get()
            ignored_files=[ metadata["source"] for metadata in collection["metadatas"]]
        else:
            print(f"Creating new vectorstore at {self.persist_directory}")
            ignored_files = []

        texts = process_documents(
            source_directory,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            ignored_files=ignored_files,
            ignore_fn=ignore_fn,
            pdf_use_unstructured=pdf_use_unstructured,
            **kwargs

        )
        self.store_documents(texts)

        if texts:
            print(
                "Ingestion complete! You can now query your documents using the LLM.ask or LLM.chat methods"
            )
        db = None
        return

# %% ../nbs/01_ingest.ipynb 6
from .utils import get_datadir

os.environ["TOKENIZERS_PARALLELISM"] = "0"
DEFAULT_DB = "vectordb"


class Ingester:
    def __init__(
        self,
        embedding_model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
        embedding_model_kwargs: dict = {"device": "cpu"},
        embedding_encode_kwargs: dict = {"normalize_embeddings": False},
        persist_directory: Optional[str] = None,
    ):
        """
        Ingests all documents in `source_folder` (previously-ingested documents are ignored)

        **Args**:

          - *embedding_model*: name of sentence-transformers model
          - *embedding_model_kwargs*: arguments to embedding model (e.g., `{device':'cpu'}`)
          - *embedding_encode_kwargs*: arguments to encode method of
                                       embedding model (e.g., `{'normalize_embeddings': False}`).
          - *persist_directory*: Path to vector database (created if it doesn't exist).
                                 Default is `onprem_data/vectordb` in user's home directory.


        **Returns**: `None`
        """
        self.persist_directory = persist_directory or os.path.join(
            get_datadir(), DEFAULT_DB
        )
        self.embeddings = HuggingFaceEmbeddings(
            model_name=embedding_model_name,
            model_kwargs=embedding_model_kwargs,
            encode_kwargs=embedding_encode_kwargs,
        )
        self.chroma_settings = Settings(
            persist_directory=self.persist_directory, anonymized_telemetry=False
        )
        self.chroma_client = chromadb.PersistentClient(
            settings=self.chroma_settings, path=self.persist_directory
        )
        return

    def get_db(self):
        """
        Returns an instance to the `langchain_chroma.Chroma` instance
        """
        db = Chroma(
            persist_directory=self.persist_directory,
            embedding_function=self.embeddings,
            client_settings=self.chroma_settings,
            client=self.chroma_client,
            collection_metadata={"hnsw:space": "cosine"},
            collection_name=COLLECTION_NAME,
        )
        return db if does_vectorstore_exist(db) else None

    def get_embedding_model(self):
        """
        Returns an instance to the `langchain_huggingface.HuggingFaceEmbeddings` instance
        """
        return self.embeddings


    def get_ingested_files(self):
        """
        Returns a list of files previously added to vector database (typically via `LLM.ingest`)
        """
        return set([d['source'] for d in self.get_db().get()['metadatas']])


    def store_documents(self, documents):
        """
        Stores instances of `langchain_core.documents.base.Document` in vectordb
        """
        if not documents: return
        db = self.get_db()
        if db:
            print(f"Creating embeddings. May take some minutes...")
            chunk_batches, total_chunks = batchify_chunks(documents)
            for lst in tqdm(chunk_batches, total=total_chunks):
                db.add_documents(lst)
        else:
            chunk_batches, total_chunks = batchify_chunks(documents)
            print(f"Creating embeddings. May take some minutes...")
            db = None

            for lst in tqdm(chunk_batches, total=total_chunks):
                if not db:
                    db = Chroma.from_documents(
                        lst,
                        self.embeddings,
                        persist_directory=self.persist_directory,
                        client_settings=self.chroma_settings,
                        client=self.chroma_client,
                        collection_metadata={"hnsw:space": "cosine"},
                        collection_name=COLLECTION_NAME,
                    )
                else:
                    db.add_documents(lst)
        return


    def ingest(
        self,
        source_directory: str, # path to folder containing document store
        chunk_size: int = DEFAULT_CHUNK_SIZE, # text is split to this many characters by [langchain.text_splitter.RecursiveCharacterTextSplitter](https://api.python.langchain.com/en/latest/character/langchain_text_splitters.character.RecursiveCharacterTextSplitter.html)
        chunk_overlap: int = DEFAULT_CHUNK_OVERLAP, # character overlap between chunks in `langchain.text_splitter.RecursiveCharacterTextSplitter`
        ignore_fn:Optional[Callable] = None, # Optional function that accepts the file path (including file name) as input and returns `True` if file path should not be ingested.
        pdf_use_unstructured:bool=False, # If True, use unstructured for PDF extraction
        **kwargs
    ) -> None:
        """
        Ingests all documents in `source_directory` (previously-ingested documents are
        ignored). When retrieved, the
        [Document](https://api.python.langchain.com/en/latest/documents/langchain_core.documents.base.Document.html)
        objects will each have a `metadata` dict with the absolute path to the file
        in `metadata["source"]`.
        Extra kwargs fed to `langchain_community.document_loaders.pdf.UnstructuredPDFLoader` when pdf_use_unstructured is True
        """

        if not os.path.exists(source_directory):
            raise ValueError("The source_directory does not exist.")
        elif os.path.isfile(source_directory):
            raise ValueError(
                "The source_directory argument must be a folder, not a file."
            )
        texts = None
        db = self.get_db()
        if db:
            # Update and store locally vectorstore
            print(f"Appending to existing vectorstore at {self.persist_directory}")
            collection = db.get()
            ignored_files=[ metadata["source"] for metadata in collection["metadatas"]]
        else:
            print(f"Creating new vectorstore at {self.persist_directory}")
            ignored_files = []

        texts = process_documents(
            source_directory,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap,
            ignored_files=ignored_files,
            ignore_fn=ignore_fn,
            pdf_use_unstructured=pdf_use_unstructured,
            **kwargs

        )
        self.store_documents(texts)

        if texts:
            print(
                f"Ingestion complete! You can now query your documents using the LLM.ask or LLM.chat methods"
            )
        db = None
        return



"""Wrappers for few-shot classification"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/04_pipelines.classifier.ipynb.

# %% auto 0
__all__ = ['DEFAULT_SETFIT_MODEL', 'SMALL_SETFIT_MODEL', 'DATASET_TEXT', 'DATASET_LABEL', 'ClassifierBase', 'FewShotClassifier']

# %% ../../nbs/04_pipelines.classifier.ipynb 3
from typing import List, Union

import warnings

with warnings.catch_warnings():
    warnings.filterwarnings("ignore",category=DeprecationWarning)
    from setfit import SetFitModel, TrainingArguments, Trainer, sample_dataset

# %% ../../nbs/04_pipelines.classifier.ipynb 4
import warnings
from abc import ABC, abstractmethod
DEFAULT_SETFIT_MODEL = "sentence-transformers/paraphrase-mpnet-base-v2"
SMALL_SETFIT_MODEL = "sentence-transformers/all-MiniLM-L6-v2"
DATASET_TEXT = "text"
DATASET_LABEL = "label"

class ClassifierBase(ABC):

  
    @abstractmethod
    def train(self,
              X:List[str],
              y:Union[List[int], List[str]],
              max_steps:int=50,
              num_epochs:int=10,
              batch_size:int=32,
              metric='accuracy',
              callbacks=None,
              **kwargs,
             ):
        """
        Trains the classifier on a list of texts (`X`) and a list of labels (`y`).
        Additional keyword arguments are passed directly to `SetFit.TrainingArguments`.

        **Args:**

        - *X*: List of texts
        - *y*: List of integers representing labels
        - *max_steps*: If set to a positive number, the total number of training steps to perform. Overrides num_epochs. 
        - *num_epochs*: Number of epochs to train
        - *batch_size*: Batch size
        - *metric*: metric to use
        - *callbacks*:  A list of callbacks to customize the training loop.

        **Returns:**

        - None
        """
        pass

    @abstractmethod
    def save(self, save_path:str):
        """
        Save model to specified folder path, `save_path`
        """
        pass


    def sample_examples(self, X:list, y:list, num_samples:int=8,
                        text_key:str=DATASET_TEXT, label_key:str=DATASET_LABEL):
        """
        Sample a dataset with `num_samples` per class
        """
        full_dataset = self.arrays2dataset(X, y, text_key=text_key, label_key=label_key)
                
        sample = sample_dataset(full_dataset, label_column=label_key, num_samples=num_samples)
        return sample.to_dict()[text_key], sample.to_dict()[label_key]
        

    
    def arrays2dataset(self, X:List[str], y:Union[List[int], List[str]], 
                       text_key:str=DATASET_TEXT, label_key:str=DATASET_LABEL):
        """
        Convert train or test examples to HF dataset
        """
        from datasets import Dataset
        return Dataset.from_dict({text_key:X, label_key:y})


    def dataset2arrays(self, dataset, text_key:str=DATASET_TEXT, label_key:str=DATASET_LABEL):
        """
        Convert a Hugging Face dataset to X, y arrays
        """
        return dataset.to_dict()['text'], dataset.to_dict()['label']
        
    def get_trainer(self):
        """
        Retrieves last trainer
        """
        if not self.trainer:
            raise ValueError('A trainer has not been created yet. You must first train a model on some labeled examples ' +\
                             'using the FewShotClassifier.train method.')
        return self.trainer

    
    def evaluate(self, X_eval:list, y_eval:list, print_report:bool=False, labels:List[str]=[]):
        """
        Evaluates labeled data using the trained model. 
        If `print_report` is True, prints classification report and returns nothing.
        Otherwise, returns and prints a dictionary of the results.

        
        """
        from sklearn.metrics import classification_report
        y_pred= self.predict(X_eval)
        if self.model.labels:
            y_pred = [self.model.labels.index(y) for y in y_pred]
            if y_eval[0] in self.model.labels:
                y_eval = [self.model.labels.index(y) for y in y_eval]

        result = classification_report(y_eval, y_pred, 
                                       output_dict=not print_report, 
                                       target_names= self.get_labels(labels))
        if print_report:
            return result
        else:
            import yaml
            print(yaml.dump(result, allow_unicode=True, default_flow_style=False))
            return result



    def explain(self, X:list, labels:List[str]=[]):
        """
        Explain the predictions on given examples in `X`. (Requires `shap` and `matplotlib` to be installed.)
        """
        try:
            import shap
        except ImportError:
            raise ImportError('Please install the shap library: pip install shap')

        try:
            import matplotlib
        except ImportError:
            raise ImportError('Please install the matplotlib library: pip install matplotlib')

        def f(x):
            return self.predict_proba(x)
        from transformers import AutoTokenizer
        tokenizer = AutoTokenizer.from_pretrained(self.model_id_or_path)

        
        output_names = self.get_labels(labels)
        explainer = shap.Explainer(f, tokenizer, output_names=self.get_labels(labels))
        shap_values = explainer(X)
        shap.plots.text(shap_values)
          
    def get_labels(self, labels:List[str]=[]):
        """
        Inspect model and return labels
        """
        target_names = labels if labels else self.model.labels
        target_names = target_names if target_names else None
        return target_names

class FewShotClassifier(ClassifierBase):
    def __init__(
        self,
        model_id_or_path:str=DEFAULT_SETFIT_MODEL,
        use_smaller:bool=False,
        **kwargs,
    ):
        """
        `FewShotClassifier` can be used to train and run text classifiers. Currently based on SetFit.
                Additional keyword arguments are fed directly to `from_pretrained`.


        **Args:**

        - *model_id_or_path*: The Hugging Face model_id or path to model folder (e.g, path previously trained and saved model).
        - *use_smaller*:  If True, will use a smaller but performant model.

        """
        self.model_id_or_path = model_id_or_path
        if use_smaller and model_id_or_path != DEFAULT_SETFIT_MODEL:
            warnings.warn(f'Over-writing supplied model ({model_id_or_path}) with {SMALL_MODEL} because use_smaller=True.')
        self.model_id_or_path = SMALL_SETFIT_MODEL if use_smaller else self.model_id_or_path
        self.model = SetFitModel.from_pretrained(self.model_id_or_path, **kwargs)
        self.predict = self.model.predict
        self.predict_proba = self.model.predict_proba
        self.trainer = None # set in `FewShotClassifier.train`
        self.labels = [] # set in `FewShotClassifier.train`
         
        
    def train(self,
              X:List[str],
              y:Union[List[int], List[str]],
              num_epochs:int=10,
              batch_size:int=32,
              metric='accuracy',
              callbacks=None,
              **kwargs,
             ):
        """
        Trains the classifier on a list of texts (`X`) and a list of labels (`y`).
        Additional keyword arguments are passed directly to `SetFit.TrainingArguments`

        **Args:**

        - *X*: List of texts
        - *y*: List of integers representing labels
        - *num_epochs*: Number of epochs to train
        - *batch_size*: Batch size
        - *metric*: metric to use
        - *callbacks*:  A list of callbacks to customize the training loop.

        **Returns:**

        - None
        """

        # convert to HF dataset
        train_dataset = self.arrays2dataset(X, y, text_key='text', label_key='label')

        args = TrainingArguments(
                batch_size=batch_size,
                num_epochs=num_epochs,
                **kwargs
        )      

        trainer = Trainer(
                    model=self.model,
                    args=args,
                    metric=metric,
                    callbacks=callbacks,
                    train_dataset=train_dataset,
                    column_mapping={"text": "text", "label": "label"}
        )
        trainer.train()

        # DEPRECATED
        # trainer = SetFitTrainer(model=self.model, train_dataset=train_dataset)
        # trainer.train(batch_size=batch_size, num_epochs=num_epochs, **kwargs)
      
        self.trainer = trainer
  

    def save(self, save_path:str):
        """
        Save model to specified folder path, `save_path`
        """
        self.model.save_pretrained(save_path)        



"""some utility functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_utils.ipynb.

# %% auto 0
__all__ = ['download', 'get_datadir', 'split_list', 'segment', 'md_to_df', 'html_to_df', 'df_to_md']

# %% ../nbs/02_utils.ipynb 3
import os.path
import requests
import sys


def download(url, filename, verify=False):
    with open(filename, "wb") as f:
        response = requests.get(url, stream=True, verify=verify)
        total = response.headers.get("content-length")

        if total is None:
            f.write(response.content)
        else:
            downloaded = 0
            total = int(total)
            # print(total)
            for data in response.iter_content(
                chunk_size=max(int(total / 1000), 1024 * 1024)
            ):
                downloaded += len(data)
                f.write(data)
                done = int(50 * downloaded / total)
                sys.stdout.write("\r[{}{}]".format("█" * done, "." * (50 - done)))
                sys.stdout.flush()


def get_datadir():
    home = os.path.expanduser("~")
    datadir = os.path.join(home, "onprem_data")
    if not os.path.isdir(datadir):
        os.mkdir(datadir)
    return datadir


def split_list(input_list, chunk_size):
    for i in range(0, len(input_list), chunk_size):
        yield input_list[i : i + chunk_size]


from syntok import segmenter
import textwrap
def segment(text:str, unit:str='paragraph', maxchars:int=2048):
    """
    Segments text into a list of paragraphs or sentences depending on value of `unit` 
    (one of `{'paragraph', 'sentence'}`. The `maxchars` parameter is the maximum size
    of any unit of text.
    """
    units = []
    for paragraph in segmenter.analyze(text):
        sentences = []
        for sentence in paragraph:
            text = ""
            for token in sentence:
                text += f'{token.spacing}{token.value}'
            sentences.append(text)
        if unit == 'sentence':
            units.extend(sentences)
        else:
            units.append(" ".join(sentences))
    chunks = []
    for s in units:
        parts = textwrap.wrap(s, maxchars, break_long_words=False)
        chunks.extend(parts)
    return chunks




from typing import Any

from io import StringIO


def md_to_df(md_str: str) -> Any:
    """Convert Markdown to dataframe."""
    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "You must install the `pandas` package to use this node parser."
        )

    # Replace " by "" in md_str
    md_str = md_str.replace('"', '""')

    # Replace markdown pipe tables with commas
    md_str = md_str.replace("|", '","')

    # Remove the second line (table header separator)
    lines = md_str.split("\n")
    md_str = "\n".join(lines[:1] + lines[2:])

    # Remove the first and last second char of the line (the pipes, transformed to ",")
    lines = md_str.split("\n")
    md_str = "\n".join([line[2:-2] for line in lines])

    # Check if the table is empty
    if len(md_str) == 0:
        return None

    # Use pandas to read the CSV string into a DataFrame
    return pd.read_csv(StringIO(md_str))


def html_to_df(html_str: str) -> Any:
    """Convert HTML to dataframe."""
    try:
        from lxml import html
    except ImportError:
        raise ImportError(
            "You must install the `lxml` package to use this node parser."
        )

    try:
        import pandas as pd
    except ImportError:
        raise ImportError(
            "You must install the `pandas` package to use this node parser."
        )

    tree = html.fromstring(html_str)
    table_element = tree.xpath("//table")[0]
    rows = table_element.xpath(".//tr")
    try:
        colnames = table_element.xpath(".//th")
        colnames = [col.text for col in colnames]
    except Exception as e:
        print(str(e))
        colnames = []

    data = []
    for row in rows:
        cols = row.xpath(".//td")
        cols = [c.text.strip() if c.text is not None else "" for c in cols]
        if len(cols) == 0: continue
        data.append(cols)

    # Check if the table is empty
    if len(data) == 0:
        return None

    # Check if the all rows have the same number of columns
    if not all(len(row) == len(data[0]) for row in data):
        return None
    if colnames:
        return pd.DataFrame(data[0:], columns=colnames)
    else:
        return pd.DataFrame(data[1:], columns=data[0])


def df_to_md(df, title=None):
	"""
	Converts pd.Dataframe to markdown
	"""
	table_md = "|"
	for col_name, col in df.items():
		table_md += f"{col_name}|"
	table_md += "\n|"
	for col_name, col in df.items():
		table_md += "---|"
	table_md += "\n"
	for row in df.itertuples():
		table_md += "|"
		for col in row[1:]:
			table_md += f"{col}|"
		table_md += "\n"

	table_summary = ""
	if title:
		table_summary = (
			f'The title of the following table in markdown format is "{title}" and '
			+ "includes this list of columns:\n"
		)
	elif title == "":
		table_summary = "The following table in markdown format includes this list of columns:\n"

	if title is not None:
		for col in df.columns:
			table_summary += f"- {col}\n"

	return table_summary + "\n" + table_md if title is not None else table_md



"""HF Pipeline"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/05_hf.hfpipeline.ipynb.

# %% auto 0
__all__ = ['HFPipeline']

# %% ../../nbs/05_hf.hfpipeline.ipynb 3
"""
Hugging Face Transformers pipeline wrapper module
"""

import inspect

from transformers import pipeline

from .models import Models

from .tensors import Tensors
from .resolver import Resolver


class HFPipeline(Tensors):
    """
    Light wrapper around Hugging Face Transformers pipeline component for selected tasks. Adds support for model
    quantization and minor interface changes.
    """

    def __init__(self, task, path=None, quantize=False, gpu=False, model=None, **kwargs):
        """
        Loads a new pipeline model.

        Args:
            task: pipeline task or category
            path: optional path to model, accepts Hugging Face model hub id, local path or (model, tokenizer) tuple.
                  uses default model for task if not provided
            quantize: if model should be quantized, defaults to False
            gpu: True/False if GPU should be enabled, also supports a GPU device id
            model: optional existing pipeline model to wrap
            kwargs: additional keyword arguments to pass to pipeline model
        """

        if model:
            # Check if input model is a Pipeline or a HF pipeline
            self.pipeline = model.pipeline if isinstance(model, HFPipeline) else model
        else:
            # Get device
            deviceid = Models.deviceid(gpu) if "device_map" not in kwargs else None
            device = Models.device(deviceid) if deviceid is not None else None

            # Split into model args, pipeline args
            modelargs, kwargs = self.parseargs(**kwargs)

            # Transformer pipeline task
            if isinstance(path, (list, tuple)):
                # Derive configuration, if possible
                config = path[1] if path[1] and isinstance(path[1], str) else None

                # Load model
                model = Models.load(path[0], config, task)

                self.pipeline = pipeline(task, model=model, tokenizer=path[1], device=device, model_kwargs=modelargs, **kwargs)
            else:
                self.pipeline = pipeline(task, model=path, device=device, model_kwargs=modelargs, **kwargs)

            # Model quantization. Compresses model to int8 precision, improves runtime performance. Only supported on CPU.
            if deviceid == -1 and quantize:
                # pylint: disable=E1101
                self.pipeline.model = self.quantize(self.pipeline.model)

        # Detect unbounded tokenizer typically found in older models
        Models.checklength(self.pipeline.model, self.pipeline.tokenizer)

    def parseargs(self, **kwargs):
        """
        Inspects the pipeline method and splits kwargs into model args and pipeline args.

        Args:
            kwargs: all keyword arguments

        Returns:
            (model args, pipeline args)
        """

        # Get pipeline method arguments
        args = inspect.getfullargspec(pipeline).args

        # Add additional pipeline args not explicitly named by pipeline function
        extra_pipeline_args = ['do_sample',
                               'max_new_tokens',
                               'repetition_penalty',
                               'return_full_text',
                               'streamer']
        args.extend(extra_pipeline_args)


        # Resolve torch dtype, if necessary
        dtype = kwargs.get("torch_dtype")
        if dtype and isinstance(dtype, str) and dtype != "auto":
            kwargs["torch_dtype"] = Resolver()(dtype)

        # Split into modelargs and kwargs
        return ({arg: value for arg, value in kwargs.items() if arg not in args}, {arg: value for arg, value in kwargs.items() if arg in args})

    def maxlength(self):
        """
        Gets the max length to use for generate calls.

        Returns:
            max length
        """

        return Models.maxlength(self.pipeline.model, self.pipeline.tokenizer)




import os
import yaml
from pathlib import Path
import mimetypes
import streamlit as st
from langchain.callbacks.base import BaseCallbackHandler
from sentence_transformers import SentenceTransformer, util
from onprem import LLM, utils as U

DATADIR = U.get_datadir()
DEFAULT_PROMPT = "List three cute names for a cat."
DEFAULT_YAML = """
llm:
  # model url (or model file name if previously downloaded)
  # if changing, be sure to update the prompt_template variable below
  model_url: https://huggingface.co/TheBloke/zephyr-7B-beta-GGUF/resolve/main/zephyr-7b-beta.Q4_K_M.gguf
  # number of layers offloaded to GPU
  n_gpu_layers: 32
  # path to vector db folder
  vectordb_path: {datadir}/vectordb
  # path to model download folder
  model_download_path: {datadir}
  # number of source documents used by LLM.ask and LLM.chat
  rag_num_source_docs: 6
  # minimum similarity score for source to be considered by LLM.ask/LLM.chat
  rag_score_threshold: 0.0
  # verbosity of Llama.cpp
  verbose: TRUE
  # additional parameters added in the "llm" YAML section will be fed directly to LlamaCpp (e.g., temperature)
  #temperature: 0.0
prompt:
  # The default prompt_template is specifically for the Zephyr-7B model.
  # It will need to be changed if you change the model_url above.
  prompt_template: <|system|>\\n</s>\\n<|user|>\\nPROMPT_VARIABLE</s>\\n<|assistant|>
ui:
  # title of application
  title: OnPrem.LLM
  # subtitle in "Talk to Your Documents" screen
  rag_title:
  # path to markdown file with contents that will be inserted below rag_title
  rag_text_path:
  # path to folder containing raw documents (i.e., absolute path of folder you supplied to LLM.ingest)
  rag_source_path:
  # base url (leave blank unless you're running your own separate web server to serve source documents)
  rag_base_url:
"""
DEFAULT_YAML_FNAME = "webapp.yml"
DEFAULT_YAML_FPATH = os.path.join(DATADIR, DEFAULT_YAML_FNAME)


def write_default_yaml():
    """
    write default webapp.yml
    """
    yaml_content = DEFAULT_YAML.format(datadir=U.get_datadir()).strip()
    yaml_content = yaml_content.replace('PROMPT_VARIABLE', '{prompt}')
    with open(DEFAULT_YAML_FPATH, "w") as f:
        f.write(yaml_content)
    return


def read_config():
    """
    Read config file.  Returns a dictionary of the configuration and a boolean indicating whether or not a new config was created.
    """
    exists = os.path.exists(DEFAULT_YAML_FPATH)
    if not exists:
        write_default_yaml()
    with open(DEFAULT_YAML_FPATH, "r") as stream:
        cfg = yaml.safe_load(stream)
    return cfg, not exists


@st.cache_resource
def load_llm():
    llm_config = read_config()[0]["llm"]
    return LLM(confirm=False, **llm_config)


@st.cache_resource
def get_embedding_model():
    model = SentenceTransformer("all-MiniLM-L6-v2")
    return model


def compute_similarity(sentence1, sentence2):
    model = get_embedding_model()
    embedding1 = model.encode(sentence1, convert_to_tensor=True)
    embedding2 = model.encode(sentence2, convert_to_tensor=True)
    cosine_score = util.pytorch_cos_sim(embedding1, embedding2)
    return cosine_score.cpu().numpy()[0][0]


class StreamHandler(BaseCallbackHandler):
    def __init__(self, container, initial_text="", display_method="markdown"):
        self.container = container
        self.text = initial_text
        self.display_method = display_method

    def on_llm_new_token(self, token: str, **kwargs) -> None:
        self.text += token + ""
        display_function = getattr(self.container, self.display_method, None)
        if display_function is not None:
            display_function(self.text)
        else:
            raise ValueError(f"Invalid display_method: {self.display_method}")


def setup_llm():
    llm = load_llm()
    chat_box = st.empty()
    stream_handler = StreamHandler(chat_box, display_method="write")
    _ = llm.load_llm()
    llm.llm.callbacks = [stream_handler]
    return llm


def check_create_symlink(source_path, base_url):
    """
    Symlink to folder named <name> in datadir from streamlit's static folder
    """
    if base_url or not source_path:
        return source_path, base_url

    # set new source path
    new_source_path = os.path.dirname(source_path)
    symlink_name = os.path.basename(source_path)

    # check for existence
    staticdir = os.path.join(os.path.dirname(st.__file__), "static")
    if os.path.islink(os.path.join(staticdir, symlink_name)):
        return new_source_path, symlink_name

    # attempt creation
    try:
        os.symlink(new_source_path, os.path.join(staticdir, symlink_name))
    except Exception:
        return source_path, base_url
    return new_source_path, base_url


def construct_link(filepath, source_path=None, base_url=None):
    """
    constructs a link to a document
    """
    import urllib

    filename = os.path.basename(filepath)
    if source_path is None:
        return filename
    base_url = base_url or "/"
    relative = str(Path(filepath).relative_to(source_path))
    link = os.path.join(base_url, relative)
    return (
        f'<a href="{urllib.parse.quote(link)}" '
        + f'target="_blank" title="Click to view original source">{filename}</a>'
    )


def is_txt(fpath):
    try:
        result = mimetypes.guess_type(fpath)
        return result[0] and result[0].startswith("text/")
    except:
        return False


def main():
    # Page setup
    cfg, cfg_was_created = read_config()

    TITLE = cfg.get("ui", {}).get("title", "OnPrem.LLM")
    RAG_TITLE = cfg.get("ui", {}).get("rag_title", None)
    APPEND_TO_PROMPT = cfg.get("prompt", {}).get("append_to_prompt", "")
    RAG_TEXT = None
    RAG_TEXT_PATH = cfg.get("ui", {}).get("rag_text_path", None)
    PROMPT_TEMPLATE = cfg.get("prompt", {}).get("prompt_template", None)
    if RAG_TEXT_PATH and os.path.isfile(RAG_TEXT_PATH) and is_txt(RAG_TEXT_PATH):
        with open(RAG_TEXT_PATH, "r") as f:
            RAG_TEXT = f.read()
    RAG_SOURCE_PATH = cfg.get("ui", {}).get("rag_source_path", None)
    RAG_BASE_URL = cfg.get("ui", {}).get("rag_base_url", None)
    RAG_SOURCE_PATH, RAG_BASE_URL = check_create_symlink(RAG_SOURCE_PATH, RAG_BASE_URL)

    st.set_page_config(page_title=TITLE, page_icon="🐍", layout="wide")
    st.title(TITLE)
    if cfg_was_created:
        st.warning(
            f"No {DEFAULT_YAML_FNAME} file was found in {DATADIR}, so a default one was created for you. Please edit as necessary."
        )

    screen = st.sidebar.radio(
        "Choose a Screen:", ("Talk to Your Documents", "Use Prompts to Solve Problems")
    )
    st.sidebar.markdown("**Curent Model:**")
    st.sidebar.markdown(f"*{os.path.basename(cfg['llm']['model_url'])}*")
    if screen == "Talk to Your Documents":
        st.sidebar.markdown(
            "**Note:** Be sure to check any displayed sources to guard against hallucinations in answers."
        )
        if RAG_TITLE:
            st.header(RAG_TITLE)
        if RAG_TEXT:
            st.markdown(RAG_TEXT, unsafe_allow_html=True)
        question = st.text_input(
            "Enter a question and press the `Ask` button:",
            value="",
            help="Tip: If you don't like the answer quality after pressing 'Ask', try pressing the Ask button a second time. "
            "You can also try re-phrasing the question.",
        )
        ask_button = st.button("Ask")
        llm = setup_llm()

        if question and ask_button:
            question = question + " " + APPEND_TO_PROMPT
            print(f'question : {question}')
            print(f'prompt_template: {PROMPT_TEMPLATE}')
            result = llm.ask(question, prompt_template=PROMPT_TEMPLATE)
            answer = result["answer"]
            docs = result["source_documents"]
            unique_sources = set()
            for doc in docs:
                answer_score = compute_similarity(answer, doc.page_content)
                question_score = compute_similarity(question, doc.page_content)
                if answer_score < 0.5 or question_score < 0.3:
                    continue
                unique_sources.add(
                    (
                        doc.metadata["source"],
                        doc.metadata.get("page", None),
                        doc.page_content,
                        question_score,
                        answer_score,
                    )
                )
            unique_sources = list(unique_sources)
            unique_sources.sort(key=lambda tup: tup[-1], reverse=True)
            if unique_sources:
                st.markdown(
                    "**One or More of These Sources Were Used to Generate the Answer:**"
                )
                st.markdown(
                    "*You can inspect these sources for more information and to also guard against hallucinations in the answer.*"
                )
                for source in unique_sources:
                    fname = source[0]
                    fname = construct_link(
                        fname, source_path=RAG_SOURCE_PATH, base_url=RAG_BASE_URL
                    )
                    page = source[1] + 1 if isinstance(source[1], int) else source[1]
                    content = source[2]
                    question_score = source[3]
                    answer_score = source[4]
                    st.markdown(
                        f"- {fname} {', page '+str(page) if page else ''} : score: {answer_score:.3f}",
                        help=f"{content}... (QUESTION_TO_SOURCE_SIMILARITY: {question_score:.3f})",
                        unsafe_allow_html=True,
                    )
            elif "I don't know" not in answer:
                st.warning(
                    "No sources met the criteria to be displayed. This suggests the model may not be generating answers directly from your documents "
                    + "and increases the likelihood of false information in the answer. "
                    + "You should be more cautious when using this answer."
                )
    else:
        prompt = st.text_area(
            "Submit a Prompt to the LLM:",
            "",
            height=100,
            placeholder=DEFAULT_PROMPT,
            help="Tip: If you don't like the response quality after pressing 'Submit', try pressing the button a second time. "
            "You can also try re-phrasing the prompt.",
        )
        submit_button = st.button("Submit")
        st.markdown(
            "*Examples of using prompts to solve different problems are [here](https://amaiya.github.io/onprem/examples.html).*"
        )
        st.markdown("---")
        llm = setup_llm()
        if prompt and submit_button:
            print(prompt)
            saved_output = llm.prompt(prompt, prompt_template=PROMPT_TEMPLATE)


if __name__ == "__main__":
    main()
